`define NUM_CORE 2

module oc_reed_solomon_decoder_top(handsome_clk, handsome_din_top, handsome_dout_top, clk, reset);
	input handsome_clk;
	input handsome_din_top;
	output handsome_dout_top;
	input clk;
	input reset;

localparam input_width = 9;
localparam output_width = 10;
localparam reducer_log4 = 2;
localparam reducer_log4_top = 4;

wire [`NUM_CORE-1:0] CE;
wire [`NUM_CORE-1:0] CEO;
wire [`NUM_CORE-1:0] Valid_out;
wire [7:0] input_byte [`NUM_CORE-1:0];
wire [7:0] Out_byte [`NUM_CORE-1:0];
wire [`NUM_CORE-1:0] handsome_din;
wire [`NUM_CORE-1:0] handsome_dout;

handsome_lfsr #(`NUM_CORE) input_lfsr_top (
			.clock(handsome_clk),
			.input_bit(handsome_din_top),
			.output_vector(handsome_din));

generate
begin:G1
	genvar i;
	for (i = 0; i < `NUM_CORE; i = i+1)
	begin :stamp_inst

		handsome_lfsr #(input_width) input_lfsr (
			.clock(handsome_clk),
			.input_bit(handsome_din[i]),
			.output_vector({CE[i],input_byte[i]}));

		oc_reed_solomon_decoder oc_stamp(
			.clk(clk),
			.reset(reset),
			.CE(CE[i]),
			.input_byte(input_byte[i]),
			.CEO(CEO[i]),
			.Valid_out(Valid_out[i]),
			.Out_byte(Out_byte[i]));

		handsome_reducer #(reducer_log4) output_reducer (
			.clock(handsome_clk),
			.input_vector({CEO[i],Valid_out[i],Out_byte[i]}),
			.output_bit(handsome_dout[i]));

	end

handsome_reducer #(reducer_log4_top) output_reducer_top (
	.clock(handsome_clk),
	.input_vector(handsome_dout[`NUM_CORE-1:0]),
	.output_bit(handsome_dout_top));

end
endgenerate
endmodule
